#include "DAQDataFormats/RawMicroBlockHeaderV0.h"
#include "DAQDataFormats/RawMilliSliceHeader.h"
#include <boost/format.hpp>

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawMicroBlockHeader
#define  CURRENT_CLASS_VERSION        0
#define  CURRENT_FIELDS_SUB_NAMESPACE rawmicroblockhdr
#include "DAQDataFormats/DataFormatLast.inc"


namespace daqdataformats{
namespace VERSION_NAMESPACE{
//------------------------------------------------------------
// Default constructor
RawMicroBlockHeader::RawMicroBlockHeader(const version_t version):
     PREVIOUS_VERSION_CLASS_NAME(version)
    ,ZEROOUT_GET_AND_SET(DestBuffID)
    ,ZEROOUT_GET_AND_SET(Partition)
    ,ZEROOUT_GET_AND_SET(V0)
    ,ZEROOUT_GET_AND_SET(Version)
    ,_DCMChan(NULL)
{}

//------------------------------------------------------------
RawMicroBlockHeader::RawMicroBlockHeader(const RawMicroBlockHeader& copyin):
     PREVIOUS_VERSION_CLASS_NAME(copyin)
    ,DEEPCOPY_GET_AND_SET(DestBuffID)
    ,DEEPCOPY_GET_AND_SET(Partition)
    ,DEEPCOPY_GET_AND_SET(V0)
    ,DEEPCOPY_GET_AND_SET(Version)
    ,_DCMChan(NULL)
{}

//------------------------------------------------------------
// Destructor
RawMicroBlockHeader::~RawMicroBlockHeader() {
    if(_DCMChan)
        delete _DCMChan;
}

//------------------------------------------------------------
// Initialize the data members
void RawMicroBlockHeader::init(){
    EXECUTE_FUNCTION(setVersion)(_version);
    PRINT_ON_DEBUG("version is set to "<<getVersion())
}

//------------------------------------------------------------------------------
uint32_t RawMicroBlockHeader::getVersion() const           { RETURN_FUNCTION(getV0)();}
bool     RawMicroBlockHeader::setVersion(const uint32_t a) { RETURN_FUNCTION(setV0)(a);}

//------------------------------------------------------------------------------
bool RawMicroBlockHeader::setAllParameters(const void* millislice_buffer){

    const RawMilliSliceHeader millislice_header(millislice_buffer);

    setGlobalDCMID(millislice_header.getGlobalDCMID());
    EXECUTE_FUNCTION(setDestBuffID)(millislice_header.getDestBuffID());
    RETURN_FUNCTION(setPartition)(millislice_header.getPartition());
}


//------------------------------------------------------------------------------
    bool RawMicroBlockHeader::setBufferSource(const void* pointer){
        _Buffer = (void*)pointer;

        /// Check if the internal buffer is too small
        if(isInternalBuffer() && _InternalBuffer.size() < rawmicroblockhdr::WORD_DCMChan){
            _InternalBuffer.assign(rawmicroblockhdr::WORD_DCMChan + 1, 0);
        }// end of checking if the internal buffer is too small

        /// Pointer to the word of the DCMChan
        const uint32_t* pp = (uint32_t*)pointer + rawmicroblockhdr::WORD_DCMChan;

        /// If it exists, set its buffer source
        if(_DCMChan) return _DCMChan->setBufferSource(pp);

        /// Since it does not exist, create a new 
        _DCMChan = new RawDCMChan(pointer); ///< version of the NanoSliceHeader equals the version of the NanoSlice

        return true;
}

//------------------------------------------------------------------------------
// Print the MicroBlock header information
void RawMicroBlockHeader::print(std::ostream& os) const{
    const uint32_t size_of_data = EXECUTE_FUNCTION(sizeofdata)();

    os<<boost::format("\nMicroBlock Header(size=%u, version=%u):") % size_of_data % _version ;
    os<<boost::format(" getDiblockId=%u")% getDiblockId();
    os<<boost::format(" DCMId=%u"       )% getDCMId();
    os<<boost::format(" DetId=%u"       )% getDetId();
    os<<boost::format(" DestBuff=%u"    )% EXECUTE_FUNCTION(getDestBuffID)();   //! Destination buffer node
    os<<boost::format(" Partition=%u"   )% EXECUTE_FUNCTION(getPartition)();    //! Partition
    os<<boost::format(" V0=%u"          )% EXECUTE_FUNCTION(getV0)();
    os<<boost::format(" Version=%u"     )% EXECUTE_FUNCTION(getVersion)();
    os<<"\n";

    printBuffer(os);

    os<<"\nMicroBlockHeader RAW 32bit words (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        os<<" ";
        printWord(i,false, os);
    }

    os<<"\nMicroBlockHeader RAW bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        os<<std::endl;
        printWord(i,true, os);
    }
}


}}// end of namespaces