#include "DAQDataFormats/RawMilliSliceHeader.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/format.hpp>

namespace daqdataformats{

//------------------------------------------------------------
// Default constructor
RawMilliSliceHeader::RawMilliSliceHeader():
     RawDAQData(0
               ,(const init_t)            &RawMilliSliceHeader::init
               ,(const sizeofdata_t)      &RawMilliSliceHeader::sizeofdata
               ,(const setBufferSource_t) &RawDAQData::setBufferSource
               ,(const print_t)           &RawMilliSliceHeader::printStandard
               ,(const readData_t)        &RawMilliSliceHeader::readDataFromBuffer)
{
    /// Perform general routines
    defaultConstructorFixedSize();
}

//------------------------------------------------------------------------------
RawMilliSliceHeader::RawMilliSliceHeader(const RawMilliSliceHeader& copy_in):
     RawDAQData(copy_in)
{
    lastOperationsInDeepCopy(copy_in);
}

//------------------------------------------------------------------------------
RawMilliSliceHeader::RawMilliSliceHeader(const void* millisliceheader_buffer):
     RawDAQData(0
               ,(const init_t)            &RawMilliSliceHeader::init
               ,(const sizeofdata_t)      &RawMilliSliceHeader::sizeofdata
               ,(const setBufferSource_t) &RawDAQData::setBufferSource
               ,(const print_t)           &RawMilliSliceHeader::printStandard
               ,(const readData_t)        &RawMilliSliceHeader::readDataFromBuffer)
{
    /// Use default Shallow copy constructor
    copy(millisliceheader_buffer);
}

//------------------------------------------------------------
// Destructor
RawMilliSliceHeader::~RawMilliSliceHeader() { }

//------------------------------------------------------------
void RawMilliSliceHeader::init(){

    /// Need to clear the MilliSlice Header every time on the init
    clear();
    /// Set version in the buffer
    //setVersion(0);

    /// Set Pad3 in the buffer do the defalut value
    setMarker(rawmillihdr::Marker_value[0]);

    setSliceSize(sizeofdata());
}

    rawmillihdr::Mode_t RawMilliSliceHeader::getMode() const { return (rawmillihdr::Mode_t) BITFIELDGET(RAWBUFF32[rawmillihdr::WORD_Mode], rawmillihdr::MASK_Mode, rawmillihdr::SHIFT_Mode);;}

//------------------------------------------------------------------------------
    uint64_t  RawMilliSliceHeader::getTimeStart()  const { 
        return getUint64_t(getTimeStartLow(), getTimeStartHigh());
    }

//------------------------------------------------------------------------------
// Setters
    bool RawMilliSliceHeader::setMode(rawmillihdr::Mode_t mode) { BITFIELDSET(RAWBUFF32[rawmillihdr::WORD_Mode], mode, rawmillihdr::MASK_Mode, rawmillihdr::SHIFT_Mode); return true;}

    bool RawMilliSliceHeader::advanceFrameSEQ()                 { return setFrameSEQ(getFrameSEQ() + 1);}
    bool RawMilliSliceHeader::resetFrameSEQ()                   { return setFrameSEQ(0); }
    bool RawMilliSliceHeader::advanceMilliSliceSEQ()            { return setMilliSliceSEQ(getMilliSliceSEQ() + 1);}
    bool RawMilliSliceHeader::addSliceSize(uint32_t size)       { return setSliceSize(getSliceSize() + size);}
    bool RawMilliSliceHeader::advanceMicrosliceNumber()         { return setMicrosliceNumber(getMicrosliceNumber() + 1);}

//------------------------------------------------------------------------------
    bool RawMilliSliceHeader::setTimeStart(const uint64_t timestamp)  {
        uint32_t lo, hi;
        getLoAndHi(timestamp, lo, hi);

        setTimeStartLow(lo);
        return setTimeStartHigh(hi);
    }


//------------------------------------------------------------------------------
// Print the MilliSlice header information
void RawMilliSliceHeader::printStandard(std::ostream& os) const {
    
    os<<std::endl<<"MilliSlice Header:";
    os<<boost::format(" Version=%u"                   )% getVersion();      //! Version number for the MilliSlice event builder
    os<<boost::format(" Marker=0x%04x"                )% getMarker();       //! MilliSlice header delimiter
    os<<boost::format(" getDiblockId=%u"              )% getDiblockId();
    os<<boost::format(" DCMId=%u"                     )% getDCMId();
    os<<boost::format(" DetId=%u"                     )% getDetId();
    os<<boost::format(" DestBuff=%u"                  )% getDestBuffID();   //! Destination buffer node
    os<<boost::format(" Partition=%u"                 )% getPartition();    //! Partition
    os<<boost::format(" TimeStartLow=%u"              )% getTimeStartLow(); //! Time start  0-31 bits for the MilliSlice
    os<<boost::format(" TimeStartHigh=%u"             )% getTimeStartHigh();//! Time start 32-55 bits for the MilliSlice
    os<<boost::format(" TimeLength=%u"                )% getTimeLength();   //! Time length of the MilliSlice
    os<<boost::format(" isComplete=%u"                )% isComplete();      //! Status word
    os<<boost::format(" isMode=%u"                    )% getMode();      //! Status word
    os<<boost::format(" isConnectionInitialization=%u")% isConnectionInitialization();      //! Status word
    os<<boost::format(" isCloseConnection=%u"         )% isCloseConnection();      //! Status word
    os<<boost::format(" isCorruptData=%u"             )% isCorruptData();      //! Status word
    os<<boost::format(" isMissingData=%u"             )% isMissingData();      //! Status word
    os<<boost::format(" SliceSEQ=%u"                  )% getMilliSliceSEQ();     //! MilliSlice sequence identifier
    os<<boost::format(" FrameSEQ=%u"                  )% getFrameSEQ();     //! Frame sequence identifier for current 5ms time slice
    os<<boost::format(" SliceSize=%u"                 )% getSliceSize();    //! The MilliSlice size in words excluding the header
    os<<boost::format(" MicroSliceNumber=%u"          )% getMicrosliceNumber(); //! Number of subframes in the millislice
    os<<boost::format(" MilliSliceIndexNumber=%u"     )% getMilliSliceIndexNumber();
    os<<boost::format(" MicroSliceTimeLength_usec=%u" )% getMicroSliceTimeLength_usec();
    os<<boost::format(" MonteCarloFlag=%u"            )% getMonteCarloFlag();
    os<<boost::format(" isCRCCalculationUsed=%u"      )% isCRCCalculationUsed();
    os<<boost::format(" isMilliSliceIndexUsed=%u"     )% isMilliSliceIndexUsed();

    const uint32_t size_of_data = sizeofdata();

    os<<"\nMilliSliceHeader RAW 32bit words (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        os<<" ";
        printWord(i, false, os);
    }

    os<<"\nMilliSliceHeader RAW bits (converted, if required, to native endian format):";
    for(uint32_t i = 0; i < size_of_data; ++i){
        os<<std::endl;
        printWord(i, true, os);
    }

}

}// end of namespace
