#include "DAQDataFormats/RawTriggerMaskV0.h"
#include "DAQDataFormats/DataFormatException.h"
#include <boost/format.hpp>

/// It is requred to
///   1. include DAQDataFormats/DataFormatFirst.inc
///   2. define CURRENT_CLASS
///   3.    define CURRENT_CLASS_VERSION for version of the class
///      OR define LAST_CLASS_VERSION    for the main structure of the class
///   4. (optional) define CURRENT_FIELDS_SUB_NAMESPACE
///   5. include DAQDataFormats/DataFormatLast.inc
/// before the declaration of the DataFormat version class in the header.
#include "DAQDataFormats/DataFormatFirst.inc"
#define  CURRENT_CLASS                RawTriggerMask
#define  CURRENT_CLASS_VERSION        0
#define  CURRENT_FIELDS_SUB_NAMESPACE triggermask
#include "DAQDataFormats/DataFormatLast.inc"

namespace daqdataformats{
namespace VERSION_NAMESPACE{

//------------------------Constructors------------------------------------------
RawTriggerMask::RawTriggerMask(const version_t version):
             PREVIOUS_VERSION_CLASS_NAME(version)
            ,ZEROOUT_GET_AND_SET(Version       )
            ,ZEROOUT_GET_AND_SET(MonteCarloFlag)
            ,ZEROOUT_GET_AND_SET(TrigMod       )
            ,ZEROOUT_GET_AND_SET(TrigMask_Low  )
            ,ZEROOUT_GET_AND_SET(TrigMask_High )
            ,ZEROOUT_GET_AND_SET(TrigMask_High2)
            ,ZEROOUT_GET_AND_SET(Prescale      )
            ,ZEROOUT_GET_AND_SET(TriggerMask)
{}

//------------------------------------------------------------------------------
RawTriggerMask::RawTriggerMask(const RawTriggerMask& copyin):
             PREVIOUS_VERSION_CLASS_NAME(copyin)
            ,DEEPCOPY_GET_AND_SET(Version       )
            ,DEEPCOPY_GET_AND_SET(MonteCarloFlag)
            ,DEEPCOPY_GET_AND_SET(TrigMod       )
            ,DEEPCOPY_GET_AND_SET(TrigMask_Low  )
            ,DEEPCOPY_GET_AND_SET(TrigMask_High )
            ,DEEPCOPY_GET_AND_SET(TrigMask_High2)
            ,DEEPCOPY_GET_AND_SET(Prescale      )
            ,DEEPCOPY_GET_AND_SET(TriggerMask)
{}

//------------------------------------------------------------------------------
RawTriggerMask::~RawTriggerMask(){}

//------------------------------------------------------------------------------
// General Initialization
void RawTriggerMask::init(){
    EXECUTE_FUNCTION(setVersion)(_version);
}

//------------------------------------------------------------------------------
uint32_t RawTriggerMask::getTriggerMask() const{


#ifdef USE_DENIS_96BIT_TRIGGER_CODE
  /********************************************************
   * Original Math is wrong.  We have 88 primary trigger  *
   * bits.  The low word provides 24 of these, the others *
   * provide 32 each                                      *
   ********************************************************/

    if     (EXECUTE_FUNCTION(getTrigMask_Low)()){
        uint32_t number = EXECUTE_FUNCTION(getTrigMask_Low)();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
            if(number & 0x1 )return ishift;
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
    else if(EXECUTE_FUNCTION(getTrigMask_High)()){
        uint32_t number = EXECUTE_FUNCTION(getTrigMask_High)();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
            if(number & 0x1)return ishift + 32;
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
    else if(EXECUTE_FUNCTION(getTrigMask_High2)()){
        uint32_t number = EXECUTE_FUNCTION(getTrigMask_High2)();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
            if(number & 0x1)return ishift + 64;
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
#endif  // USE_DENIS_96BIT_TRIGGER_CODE

    // TMOD is used as the "page" for the triggers.  This is just a multiplication of the
    // trigger bit by the TMOD value

    uint8_t tmod = (uint8_t)(EXECUTE_FUNCTION(getTrigMod)() & 0xFF);

    if     (EXECUTE_FUNCTION(getTrigMask_Low)()){
        uint32_t number = EXECUTE_FUNCTION(getTrigMask_Low)();
        for(uint32_t ishift = 0; ishift < 24; ++ishift){
	  if(number & 0x1 ){
	    return ishift * (tmod+1);
	  }
            number = number >> 1;
        } // endfor
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
    else if(EXECUTE_FUNCTION(getTrigMask_High)()){
        uint32_t number = EXECUTE_FUNCTION(getTrigMask_High)();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
	  if(number & 0x1){
	    return (ishift + 24)*(tmod+1);
	  }
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }
    else if(EXECUTE_FUNCTION(getTrigMask_High2)()){
        uint32_t number = EXECUTE_FUNCTION(getTrigMask_High2)();
        for(uint32_t ishift = 0; ishift < 32; ++ishift){
	  if(number & 0x1){
	    return (ishift + 56)*(tmod+1);
	  }
            number = number >> 1;
        }
        std::string msg("Something is wrong in RawTriggerMask::getTriggerMask()");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return 0;
    }


    //std::string msg("From RawTriggerMask::getTriggerMask(). TriggerMask is not set");
    //GENERATE_DATAFORMAT_EXCEPTION(msg);
    return 0;
}

//------------------------------------------------------------------------------
bool RawTriggerMask::setTriggerMask(const uint32_t mask){

    uint32_t mask_copy = mask;
    uint8_t tmod;

    // Calculate the trigger page that we should be on
    // This is the bit number div the size of the pages (88 bits)
    if(mask_copy > 88){
      tmod = (mask_copy / 88);              // bit div page size
      EXECUTE_FUNCTION(setTrigMod)(tmod);   // Set the page value
      mask_copy = mask_copy - (tmod * 88);  // Subtract off the page offsets
    }


    if (mask_copy > 88){
        std::string msg("The mask is too high");
        GENERATE_DATAFORMAT_EXCEPTION(msg);
        return false;
    }

    if(mask_copy < 24){
        uint32_t number = 0x1;
        for(uint32_t ishift = 0; ishift < mask_copy; ++ishift){
            number = number << 1;
        }
        RETURN_FUNCTION(setTrigMask_Low)(number & 0x00FFFFFF);
    }
    else if(mask_copy < 56){
        mask_copy -= 24;
        uint32_t number = 0x1;
        for(uint32_t ishift = 0; ishift < mask_copy; ++ishift){
            number = number << 1;
        }
        RETURN_FUNCTION(setTrigMask_High)(number);
    }

    
        mask_copy -= 56;
        uint32_t number = 0x1;
        for(uint32_t ishift = 0; ishift < mask_copy; ++ishift){
            number = number << 1;
        }
        RETURN_FUNCTION(setTrigMask_High2)(number);
}

//------------------------------------------------------------------------------
void RawTriggerMask::print(std::ostream& os) const{

    os<<"++++ Trigger Mask ++++\n";
    os<<boost::format("Trigger Mask Version:\t%u\n") % EXECUTE_FUNCTION(getVersion)();
    os<<boost::format("Monte Carlo:\t\t%s\n"       ) %(EXECUTE_FUNCTION(getMonteCarloFlag)()? "True" : "False");
    os<<boost::format("Trigger Modification:\t0x%02x\n") % EXECUTE_FUNCTION(getTrigMod)();
    os<<boost::format("Trigger Types low:\t0x%08x\n"   ) % EXECUTE_FUNCTION(getTrigMask_Low)();
    os<<boost::format("Trigger Types mid:\t0x%08x\n"   ) % EXECUTE_FUNCTION(getTrigMask_High)();
    os<<boost::format("Trigger Types high:\t0x%08x\n"  ) % EXECUTE_FUNCTION(getTrigMask_High2)();
    os<<boost::format("Trigger Type:\t\t%u\n"      ) % EXECUTE_FUNCTION(getTriggerMask)();
    os<<boost::format("Trigger prescale:\t%u\n"    ) % EXECUTE_FUNCTION(getPrescale)();
    PRINT_ON_DEBUG("Buffer is set to "<<_Buffer)
}


}}
