#include "DatabaseUtils/DAQConfig/GlobalNameSpec.h"
#include "DatabaseUtils/DAQConfig/NamedConfigUtils.h"
#include <NovaDAQUtilities/EnvVarCache.h>
#include "NovaDatabase/Table.h"
#include "NovaDatabase/Util.h"

using namespace nova::database;
using namespace dbutils::daqconfig;

/**
 * Command-line utility for displaying named global configuration information.
 */

int main(int argc, char **argv)
{
    // initialization
    int nRow;
    std::string fname;
    boost::shared_ptr<Table> dbt;
    Row* rowPtr;
    Column* columnPtr;

    // parse command line arguments and print usage hints, if needed.
    std::string timestamp = "";
    bool printUsage = false;
    bool includeConfigsWithInvalidStates = false;
    int c;
    while ((c = getopt (argc, argv, "t:hi")) != -1) {
        switch (c)
        {
        case 't':
            timestamp = optarg;
            break;
        case 'h':
            printUsage = true;
            break;
        case 'i':
            includeConfigsWithInvalidStates  = true;
            break;
        default:
            printUsage = true;
            break;
        }
    }

    // fetch the command line arguments
    std::string daqOrSlowControls;
    int64_t globalConfigId = 0;
    std::string configName;
    std::string dcmOpMode;
    std::string detectorName;
    if (! printUsage) {
        if ((argc - optind) < 2) {
            printUsage = true;
        }
        else {
            {
                std::string tmpString(argv[optind]);
                if (tmpString == "DAQ" || tmpString == "SlowControls") {
                    daqOrSlowControls = tmpString;
                }
                else {
                    printUsage = true;
                }
                ++optind;
            }
            try {
                std::string firstArg(argv[optind]);
                globalConfigId = boost::lexical_cast<int64_t>(firstArg);
                ++optind;
            }
            catch (...) {
                if ((argc - optind) < 3) {
                    printUsage = true;
                }
                else {
                    {
                        std::string tmpString(argv[optind]);
                        configName = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        dcmOpMode = tmpString;
                        ++optind;
                    }
                    {
                        std::string tmpString(argv[optind]);
                        detectorName = tmpString;
                        ++optind;
                    }
                }
            }
        }
    }

    // print usage hints, if needed.
    if (printUsage) {
        std::cout << "Usage: " << argv[0] << " [-t '<timestamp>'] [-i] "
                  << "<\"DAQ\"|\"SlowControls\"> <globalCfgName> <dcmOpMode> "
                  << "<detectorName>" << std::endl;
        std::cout << " OR    " << argv[0] << " "
                  << "<\"DAQ\"|\"SlowControls\"> <globalCfgId>" << std::endl;
        std::cout << "where \"DAQ\"|\"SlowControls\" selects which set "
                  << "of database tables to use" << std::endl;
        std::cout << "      globalCfgName specifies the named "
                  << "configuration to show" << std::endl;
        std::cout << "      dcmOpMode specifies the DCM operating mode "
                  << "(e.g. DCS)" << std::endl;
        std::cout << "      detectorName specifies the detector name "
                  << "(e.g. NDOS)" << std::endl;
        std::cout << "      globalCfgId specifies the "
                  << "named global configuration to show" << std::endl;
        std::cout << "     -t specifies a particular time to select (UTC)"
                  << std::endl;
        std::cout << "     -i indicates that configurations should be "
                  << "fetched independent" << std::endl;
        std::cout << "        of whether they are obsolete, changing, "
                  << "or incomplete" << std::endl;
        return -1;
    }

    // fetch the existing combinations of subsystem and configtype from
    // the named subsystem configurations table
    std::vector<std::pair<std::string, std::string> > subsystemList;
    if (daqOrSlowControls == "DAQ") {
        fname = "NovaDatabase/tables/DAQConfig/NamedSubsystemConfigurations.xml";
    }
    else {
        fname = "NovaDatabase/tables/DCS/DCSNamedSubsystemConfigurations.xml";
    }
    dbt.reset(new Table(fname.c_str()));
    std::string detName = 
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADAQ_ENVIRONMENT");
    dbt->SetDetector(detName);
    dbt->GetConnection();
    if (! dbt->ExistsInDB()) {
        std::cerr << "*** ERROR: The table described by " << fname
                  << " does not exist in the database! "
                  << "(File = " << __FILE__ << ", Line = " << __LINE__
                  << ")" << std::endl;
        exit(2);
    }
    dbt->Clear();
    //dbt->SetVerbosity(1);
    dbt->AddDistinctColumn(NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME);
    dbt->AddDistinctColumn(NamedConfigUtils::CONFIGTYPE_COLUMN_NAME);
    dbt->LoadFromDB();
    nRow = dbt->NRow();
    //std::cout << "NRow = " << nRow << std::endl;
    for (int idx = 0; idx < nRow; ++idx) {
        rowPtr = dbt->GetRow(idx);
        columnPtr = rowPtr->Col(NamedConfigUtils::SUBSYSTEMNAME_COLUMN_NAME);
        std::string subsystemName;
        if (! columnPtr->Get(subsystemName)) {
            std::cerr << "*** ERROR: Unable to fetch the subsystemName "
                      << "from the table described by " << fname
                      << "! " << "(File = " << __FILE__ << ", Line = "
                      << __LINE__ << ")" << std::endl;
            exit(2);
        }
        columnPtr = rowPtr->Col(NamedConfigUtils::CONFIGTYPE_COLUMN_NAME);
        std::string configType;
        if (! columnPtr->Get(configType)) {
            std::cerr << "*** ERROR: Unable to fetch the configType "
                      << "from the table described by " << fname
                      << "! " << "(File = " << __FILE__ << ", Line = "
                      << __LINE__ << ")" << std::endl;
            exit(2);
        }
        subsystemList.push_back(std::make_pair(subsystemName, configType));
    }

    // tell the user which database he/she is using
    std::string dbName =
        novadaq::EnvVarCache::getInstance().getEnvVar("NOVADBNAME");
    std::cout << "*** You are currently using the " << dbName
              << " database. ***" << std::endl;

    // fetch the details of the user-specified global configuration
    // along with all of its ancestors
    std::vector<NamedConfigUtils::NamedConfigDetailsPtr> detailsList;
    if (globalConfigId != 0) {
        IDSpec idSpec(globalConfigId, IDSpec::GLOBAL,
                      (daqOrSlowControls == "DAQ" ? IDSpec::DAQ :
                       IDSpec::SLOWCONTROLS));
        detailsList =
            NamedConfigUtils::getConfigDetailAncestry(idSpec);
    }
    else if (timestamp.length() < 14) {
        GlobalNameSpec globalNameSpec(configName, dcmOpMode, detectorName);
        globalNameSpec.isDAQConfiguration(daqOrSlowControls == "DAQ");
        globalNameSpec.skipInvalidStateTest(includeConfigsWithInvalidStates);
        detailsList =
            NamedConfigUtils::getGlobalConfigDetailAncestry(globalNameSpec);
    }
    else {
        GlobalNameSpec globalNameSpec(configName, dcmOpMode, detectorName,
                                      timestamp);
        globalNameSpec.isDAQConfiguration(daqOrSlowControls == "DAQ");
        globalNameSpec.skipInvalidStateTest(includeConfigsWithInvalidStates);
        detailsList =
            NamedConfigUtils::getGlobalConfigDetailAncestry(globalNameSpec);
    }
    if (detailsList.size() == 0) {
        std::cerr << "*** ERROR: Unable to fetch the database IDs "
                  << "for the specified input parameters! "
                  << "(File = " << __FILE__ << ", Line = "
                  << __LINE__ << ")" << std::endl;
        exit(2);
    }

    // loop through the global configurations
    std::vector<std::pair<std::string, std::string> >
        subsystemConfigsAlreadyFound;
    std::string indent = "";
    for (uint32_t idx = 0; idx < detailsList.size(); ++idx) {
        NamedConfigUtils::NamedConfigDetailsPtr detailsPtr = detailsList[idx];
        std::string idString;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              ID_COLUMN_NAME,
                                              idString)) {
            idString = "Unknown ID";
        }
        std::string configName;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              NAME_COLUMN_NAME,
                                              configName)) {
            configName = "Unknown Name";
        }
        std::string insertTime;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              INSERTTIME_COLUMN_NAME,
                                              insertTime)) {
            insertTime = "Unknown Time";
        }
        std::string isObsolete;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              ISOBSOLETE_COLUMN_NAME,
                                              isObsolete)) {
            isObsolete = "-";
        }
        std::string isChanging;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              ISCHANGING_COLUMN_NAME,
                                              isChanging)) {
            isChanging = "-";
        }
        std::string isComplete;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              ISCOMPLETE_COLUMN_NAME,
                                              isComplete)) {
            isComplete = "-";
        }
        std::string description;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              DESCRIPTION_COLUMN_NAME,
                                              description)) {
            description = "";
        }
        std::string usageCount;
        if (! NamedConfigUtils::getFieldValue(detailsPtr,
                                              NamedConfigUtils::
                                              USAGECOUNT_COLUMN_NAME,
                                              usageCount)) {
            usageCount = "0";
        }
        std::cout << indent << "* " << configName
                  << ", id = " << idString
                  << ", time = '" << insertTime << "' (UTC)"
                  << ", flags = " << isObsolete << isChanging << isComplete
                  << std::endl;
        std::cout << indent << "      description = '"
                  << description << "', usage count = " << usageCount
                  << std::endl;


        std::string globalConfigId;
        if (NamedConfigUtils::getFieldValue(detailsPtr,
                                            NamedConfigUtils::ID_COLUMN_NAME,
                                            globalConfigId)) {
            dbt->Clear();
            //dbt->SetVerbosity(1);
            dbt->SetValidityRange(NamedConfigUtils::GLOBALCFGID_COLUMN_NAME,
                                  globalConfigId);
            dbt->LoadFromDB();
            nRow = dbt->NRow();
            //std::cout << "NRow = " << nRow << std::endl;
            for (int idx = 0; idx < nRow; ++idx) {
                rowPtr = dbt->GetRow(idx);
                std::string subsystemName;
                columnPtr = rowPtr->Col(NamedConfigUtils::
                                        SUBSYSTEMNAME_COLUMN_NAME);
                if (! columnPtr->Get(subsystemName)) {
                    subsystemName = "Unknown Subsystem";
                }
                std::string configType;
                columnPtr = rowPtr->Col(NamedConfigUtils::
                                        CONFIGTYPE_COLUMN_NAME);
                if (! columnPtr->Get(configType)) {
                    configType = "Unknown ConfigType";
                }

                bool alreadyDisplayed = false;
                for (uint32_t kdx = 0;
                     kdx < subsystemConfigsAlreadyFound.size();
                     ++kdx) {
                    if (subsystemConfigsAlreadyFound[kdx].first ==
                        subsystemName &&
                        subsystemConfigsAlreadyFound[kdx].second ==
                        configType) {
                        alreadyDisplayed = true;
                        break;
                    }
                }

                if (! alreadyDisplayed) {
                    subsystemConfigsAlreadyFound.
                        push_back(std::make_pair(subsystemName, configType));

                    columnPtr = rowPtr->Col(NamedConfigUtils::
                                            ID_COLUMN_NAME);
                    if (! columnPtr->Get(idString)) {
                        idString = "Unknown ID";
                    }
                    columnPtr = rowPtr->Col(NamedConfigUtils::
                                            NAME_COLUMN_NAME);
                    if (! columnPtr->Get(configName)) {
                        configName = "Unknown Name";
                    }
                    columnPtr = rowPtr->Col(NamedConfigUtils::
                                            INSERTTIME_COLUMN_NAME);
                    if (! columnPtr->Get(insertTime)) {
                        insertTime = "Unknown Time";
                    }
                    columnPtr = rowPtr->Col(NamedConfigUtils::
                                            DESCRIPTION_COLUMN_NAME);
                    if (! columnPtr->Get(description)) {
                        description = "";
                    }
                    columnPtr = rowPtr->Col(NamedConfigUtils::
                                            USAGECOUNT_COLUMN_NAME);
                    if (columnPtr->IsNull() || ! columnPtr->Get(usageCount)) {
                        usageCount = "0";
                    }
                    std::cout << indent << "  - " << subsystemName
                              << " " << configType << " " << configName
                              << ", id = " << idString
                              << ", time = '" << insertTime
                              << "' (UTC)" << std::endl;
                    std::cout << indent << "        description = '"
                              << description << "', usage count = " << usageCount
                              << std::endl;

                    try {
                        int64_t subsystemConfigId =
                            boost::lexical_cast<int64_t>(idString);
                        IDSpec subsystemIdSpec(subsystemConfigId,
                                               IDSpec::SUBSYSTEM,
                                               (daqOrSlowControls == "DAQ" ?
                                                IDSpec::DAQ :
                                                IDSpec::SLOWCONTROLS));
                        std::vector<NamedConfigUtils::NamedConfigDetailsPtr>
                            subsysDetailsList = NamedConfigUtils::
                            getConfigDetailAncestry(subsystemIdSpec);

                        std::string indent2 = indent;
                        for (uint32_t sdx = 1; sdx < subsysDetailsList.size(); ++sdx) {
                            indent2.append(" ");
                            NamedConfigUtils::NamedConfigDetailsPtr
                                subsysDetailsPtr = subsysDetailsList[sdx];
                            if (! NamedConfigUtils::getFieldValue(subsysDetailsPtr,
                                                                  NamedConfigUtils::
                                                                  ID_COLUMN_NAME,
                                                                  idString)) {
                                idString = "Unknown ID";
                            }
                            if (! NamedConfigUtils::getFieldValue(subsysDetailsPtr,
                                                                  NamedConfigUtils::
                                                                  NAME_COLUMN_NAME,
                                                                  configName)) {
                                configName = "Unknown Name";
                            }
                            if (! NamedConfigUtils::getFieldValue(subsysDetailsPtr,
                                                                  NamedConfigUtils::
                                                                  INSERTTIME_COLUMN_NAME,
                                                                  insertTime)) {
                                insertTime = "Unknown Time";
                            }
                            if (! NamedConfigUtils::getFieldValue(subsysDetailsPtr,
                                                                  NamedConfigUtils::
                                                                  DESCRIPTION_COLUMN_NAME,
                                                                  description)) {
                                description = "";
                            }
                            if (! NamedConfigUtils::getFieldValue(subsysDetailsPtr,
                                                                  NamedConfigUtils::
                                                                  USAGECOUNT_COLUMN_NAME,
                                                                  usageCount)) {
                                usageCount = "0";
                            }
                            std::cout << indent2 << "  - " << subsystemName
                                      << " " << configType << " " << configName
                                      << ", id = " << idString
                                      << ", time = '" << insertTime
                                      << "' (UTC)" << std::endl;
                            std::cout << indent2 << "        description = '"
                                      << description << "', usage count = " << usageCount
                                      << std::endl;
                        }
                    }
                    catch (...) {}
                }
            }
        }
        else {
            std::cout << indent << "  ERROR looking up subsystem "
                      << "configurations." << std::endl;
        }

        indent.append(" ");
    }

    return 0;
}
